// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// 
// Module: PositionUtils
// Author: Hassan Siddiqui
//
// Description
// Provides utilites for coordinate conversions,
// source position
// and CCD number determination, based on a position.
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: PositionUtils.cc,v 1.1 2011/01/11 16:36:10 rsaxton Exp $

/* General */

/* SAS-specific classes */
#include "Xmm.h"
#include "CalCommon.h"
#include "CalServer.h"
#include "OalServer.h"
#include "Tracer.h"

/* Local classes */
#include "CamCoords.h"
#include "TelCoords.h"
#include "PixCoords.h"
#include "PositionUtils.h"


/** Constructors  */
PositionUtils::PositionUtils( const Xmm::Instrument & instrumentID)
{
  PROC;

  /***************************************************
   * Determine, in PIXCOORD1, 
   * boundaries for CCDs of this detectori, taking into account 
   * the windowing of the central chip.
   * This information is used by setCcd(CamCoords)
   ***************************************************/
  
  const XmmInstrument & instr 
    = xmmPayLoad.instrument(instrumentID);

  _nccds    = instr.nCcds();
  _ccd_xmin = instr.minX();
  _ccd_xmax = instr.maxX();
  _ccd_ymin = instr.minY();
  _ccd_ymax = instr.maxY();
  _cent_ccd = instr.focusCcd();

 /* Set the extent of the observing window */
  ModeParameters *modeParam = 0; modeParam = calServer.getAtom(modeParam);
  _ccdwindow = new CCDwindow(modeParam->windowX0(), modeParam->windowY0(),
                        (modeParam->windowX0()+modeParam->windowDX()-1),
                        (modeParam->windowY0()+modeParam->windowDY()-1));

  /* Get focal length */
  MiscDataServer * mds = 0; mds = calServer.getAtom(mds);

  double focalLength = mds->get("FOCAL_LENGTH");

  /* Conversion factor between det coords (0.05 arcsec) 
     and camcoord units (mm)*/
  _detToCamcoord2 = focalLength * M_PI / (180. * 3600 * 20);

  /* Get the optical axis position for this detector 
     Not actually necessary but leave the code anyway */

  double ccdid = mds->get("OPTICS_CCD");
  double opticsx = mds->get("OPTICS_X");
  double opticsy = mds->get("OPTICS_Y");
  double plateX = mds->get("PLATE_SCALE_X");
  double plateY = mds->get("PLATE_SCALE_Y");

  //  Set the CCD node to the optical axis value 
  calServer.state()->ccd()->set(UNT(ccdid));

  //  Find the centre of the CAMCOORD2 frame in RAW pixels 
  GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);
  real64 xmm = 0.0;
  real64 ymm = 0.0;
  real64 zmm = 0.0;
  CalReal64Vector x(&xmm, 1), y(&ymm, 1), z(&zmm, 1);

  XmmCartesianPointList camCoord2 = ct->camCoord2List(x, y, z);
  XmmCartesianPointList camCoord1 = camCoord2.toCamCoord1();
  XmmPixelList chipcoord = camCoord1.toChipCoord();

  //  Find offset from optical axis and convert to det coords 
  double pixSize = 0.05;
  //_cenDetX = ( (chipcoord.x()[0]-opticsx) * plateX / pixSize + 0.5);
  //_cenDetY = ( (chipcoord.y()[0]-opticsy) * plateY / pixSize + 0.5);
  _cenDetX = ( opticsx * plateX / pixSize + 0.5);
  _cenDetY = ( opticsy * plateY / pixSize + 0.5);

    errstr<<"det centre: "<<_cenDetX<<" "<<_cenDetY
      << Sas::message(Msg::App, Msg::Verbose);

  _cenDetX = ( (chipcoord.x()[0]) * plateX / pixSize + 0.5);
  _cenDetY = ( (chipcoord.y()[0]) * plateY / pixSize + 0.5);
    errstr<<"det centre: "<<_cenDetX<<" "<<_cenDetY
      << Sas::message(Msg::App, Msg::Verbose);

   _cenDetX = 0.0;
   _cenDetY = 0.0;

  ENDPROC;
}

/*  Destructors  */
PositionUtils::~PositionUtils(){};

/* Convert from DETX/DETY (0.05 arcsec units) to camcoord2 */
const CamCoords PositionUtils::detxy005ToCamCoord2( double detx, 
						    double dety ) const
{
  PROC;
  
  /* Return a pseudo-camcoord2 object, 
     as the z property is of no relevance. */
  RETURN CamCoords( ( detx - _cenDetX ) * _detToCamcoord2, 
		    ( dety - _cenDetY ) * _detToCamcoord2,
		    1);
      
}

/* Convert from camcoord2 to DETX/DETY (0.05 arcsec units) */
const double PositionUtils::CamCoord2Todetxy005( double incamcoord ) const
{
  PROC;
  
  /* Return the camcoord2 value scaled to a detector pixel size */ 
    double detcoord = incamcoord / _detToCamcoord2;

  RETURN detcoord;
}

/* Convert from camcoord2 to DETX/DETY (0.05 arcsec units) */
const CoordPair PositionUtils::CamCoord2Todetxy005( const CamCoords & camCoord2 ) const
{
  PROC;
  
  /* Convert the camcoord2 position to detector coordinates */
    double detx = _cenDetX + camCoord2.x() / _detToCamcoord2;
    double dety = _cenDetY + camCoord2.y() / _detToCamcoord2;

  RETURN CoordPair(detx, dety);
}

/* Convert from RAW coords to DETX/DETY */
const CoordPair PositionUtils::RawTodetxy005( double rawx, double rawy, int ccdnr ) const
{
  PROC;
  
  /* Create PixCoord object from RAW position */
    PixCoords pixCoord1(UNT(rawx), UNT(rawy));
    CamCoords camCoord2 = PixCoordTocamCoord2(pixCoord1, ccdnr);
    CoordPair detPos = CamCoord2Todetxy005(camCoord2);

    errstr<<"RawToDet: "<<rawx<<" "<<rawy<<" "<<detPos.x()<<" "<<detPos.y()
      <<" CCD: "<<calServer.state()->ccd()->get()
      << Sas::message(Msg::App, Msg::Verbose);

  RETURN detPos;
}

/* Convert from CAMCOORD2 to TELCOORD */
const TelCoords PositionUtils::camCoord2ToTelCoord( const CamCoords & camCoords ) const
{
    PROC;
    /* Load up result into focal plane position object */
    GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

    real64 xmm = camCoords.x();
    real64 ymm = camCoords.y();
    real64 zmm = camCoords.z();

    CalReal64Vector x(&xmm, 1), y(&ymm, 1), z(&zmm, 1);

    /* Generate CamCoord2 object and perform conversion,
       which requires a number of steps to reach target system */
    XmmCartesianPointList camcoord2l = ct->camCoord2List(x, y, z);
    XmmTelescopePointList telcoordl  = camcoord2l.toTelCoord();
    
    double theta = telcoordl.theta()[0];
    double phi   = telcoordl.phi()[0];

    RETURN TelCoords (theta, phi);
}

/* Convert from TELCOORD to CamCoord2*/
const CamCoords PositionUtils::TelCoordTocamCoord2( const TelCoords & telCoords ) const
{
    PROC;
      
    /* Load up result into focal plane position object */
    GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

    real64 thetal = telCoords.theta();
    real64 phil = telCoords.phi();

    CalReal64Vector t(&thetal, 1), p(&phil, 1);
    XmmCartesianPointList camcoord2l = ct->telCoordList(t, p).
                                    toCamCoord2();
    double x = camcoord2l.x()[0];
    double y = camcoord2l.y()[0];

    RETURN CamCoords(x,y,1);
}

/* Convert from CAMCOORD2 to PIXCOORD */
const PixCoords PositionUtils::camCoord2ToPixCoord( const CamCoords & camCoords ) const
{
  PROC; 

     /* Load up result into CAL camcoord object */
    GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

    real64 xmm = camCoords.x();
    real64 ymm = camCoords.y();
    real64 zmm = camCoords.z();

    CalReal64Vector x(&xmm, 1), y(&ymm, 1), z(&zmm, 1);

    /* Generate CamCoord2 object and perform conversion,
       which requires a number of steps to reach target system */
    XmmCartesianPointList camcoord2l = ct->camCoord2List(x, y, z);
    XmmCartesianPointList camcoord1l = camcoord2l.toCamCoord1();
    XmmPixelList chipcoordl          = camcoord1l.toChipCoord();

    //    XmmPixelList pixcoord1l          = chipcoordl.toPixCoord1();
    /* WILL RETURN CHIPCOORDS */
    unsigned xpix = chipcoordl.x()[0];
    unsigned ypix = chipcoordl.y()[0];

    RETURN PixCoords (xpix, ypix);
}

/* Convert from PIXCOORD to CamCoord2 */
const CamCoords PositionUtils::PixCoordTocamCoord2( const PixCoords & pixCoords, int ccdnr) const
{
  PROC; 

  // if ccdnr less than 0, this routine assumes that the chip number has been preset in the calserver 
     
  /* Set up a one element chip coordinate list */
    CalInt16Vector chipX(0, 1), chipY(0, 1);
    chipX[0] = 10;
    chipY[0] = 10;
    chipX[0] = pixCoords.x();
    chipY[0] = pixCoords.y();
// dont understand this bit //
    //float xmm[NUM_OF(chipX)], ymm[NUM_OF(chipX)], zmm[NUM_OF(chipX)];

//    calServer.pushState();	// save the current calserver state

    if (ccdnr > 0) calServer.state()->ccd()->set(ccdnr);	// set correct ccd 

    GeometryDataServer *cd = 0; cd = calServer.getAtom(cd);

    XmmPixelList chipCoord1 = cd->pixCoord1List(chipX, chipY).toChipCoord();

//    calServer.popState();	// restore the calserver state

   // XmmCartesianPointList camCoord1 = chipCoord1.toCamCoord1(xmm, ymm, zmm);
    XmmCartesianPointList camCoord1 = chipCoord1.toCamCoord1();
    XmmCartesianPointList camCoord2 = camCoord1.toCamCoord2();

  RETURN CamCoords(camCoord2.x()[0], camCoord2.y()[0], 1);

}

/* Find the width in detector coordinates of a single raw pixel */
const vector<double> PositionUtils::detxyWidth( ) const
{
  PROC;
  
   /* Find the distance in detector coordinates between two adjacent pixels */
   PixCoords pixCoord1(10,10);
   CamCoords camCoord2 = PixCoordTocamCoord2(pixCoord1);

   PixCoords pixCoord1_2(11,11);
   CamCoords camCoord2_2 = PixCoordTocamCoord2(pixCoord1_2);

   vector<double> Widths;
   Widths.push_back(abs((camCoord2_2.x()-camCoord2.x()) / _detToCamcoord2));
   Widths.push_back(abs((camCoord2_2.y()-camCoord2.y()) / _detToCamcoord2));

    errstr<<"Xwidth: "<<Widths[0]<<" Ywidth: "<<Widths[1]
      << Sas::message(Msg::App, Msg::Verbose);

   RETURN (Widths);
}

/* Set ccd state according to CamCoords input */
bool PositionUtils::setCcd(const CamCoords & camcoord)
{
  PROC; 

  bool foundccd = false;

//    errstr 
//      << "setCcd: nccds " << _nccds
//      << Sas::message(Msg::App, Msg::Verbose);

  for (unsigned ccdid = 1; ccdid <= _nccds; ++ccdid)
    {
      errstr 
	<< "setCcd: set ccd to " << ccdid
	<< Sas::message(Msg::App, Msg::Noisy);
      
      calServer.state()->ccd()->set(ccdid);
      
      PixCoords pixcoord = camCoord2ToPixCoord(camcoord);
      
   /* Test if the pixel lies within this CCD taking into account the
      window mode */
      if ( (ccdid != _cent_ccd && 
               pixcoord.x() >= _ccd_xmin && pixcoord.x() <= _ccd_xmax &&
	           pixcoord.y() >= _ccd_ymin && pixcoord.y() <= _ccd_ymax )
        || (ccdid == _cent_ccd &&
               _ccdwindow->isPixelInside(
                        PixelLocation(pixcoord.x(),pixcoord.y()))) )
	{
	  
	  errstr
	    << "[pixcoord] pixcoord = ["
	    << pixcoord.x()
	    << "," 
	    << pixcoord.y()
	    << "]"
	    << Sas::message(Msg::App, Msg::Noisy);

	  foundccd = true;
	  break;
	}
      else
	{
	  errstr 
	    << "[outsideCCD] setCcd: pixcoord for " << ccdid
	    << " is [" 
	    << pixcoord.x() << ","
	    << pixcoord.y() 
	    << "] but is OUTSIDE ccd range (x="<< _ccd_xmin
	    << " to " << _ccd_xmax
	    << ", y="<< _ccd_ymin
	    << " to " << _ccd_ymax
	    << "). Moving onto next CCD"
	    << Sas::message(Msg::App, Msg::Noisy);
	}
    }

  if (!foundccd)
    {
      errstr 
	<< "[noCCDFound] PositionUtils: Could not find CCD corresponding to CAMCOORDS["
	<< camcoord.x() << ","
	<< camcoord.y() << ","
	<< camcoord.z() << "]."
	<< Sas::message(Msg::App, Msg::Noisy);

      /* Re-set CAL state */
      calServer.state()->ccd()->set(1);
    }
  
  RETURN foundccd;
}

const SourcePosition 
PositionUtils::sourcePosition( const string & sourceCoordType,
			      const string & datamode,
			      const string & dateobs,
			      double sourcex,
			      double sourcey,
			      const PosWCS * const poswcs,
			      int ccdnr)
{
  PROC;

  /**********************************
   * Get source position
   **********************************/

  /* Make code case-insensitive */
  string sctype = uppercase(sourceCoordType);

  if ( (sctype == "EQPOS") || (sctype == "POS") || (sctype == "TEL") )
    {
      double ra=0;
      double dec=0;
      real64 theta=0, phi=0;

      if  (sctype == "EQPOS")
	{
	  errstr 
	    << "[sourcepos] using input equ coords [ra="
	    << sourcex
	    << ",dec="
	    << sourcey
	    <<"] for source"
	    << Sas::message(Msg::App, Msg::Noisy);

	  ra = sourcex;
	  dec = sourcey;
	}
      else if (sctype == "POS") 
	{
         errstr
            << "Converting from "<<sourcex<<" "<<sourcey
            << Sas::message(Msg::App, Msg::Noisy);

	  /* Convert from sky [X/Y] coords 
	     to RA/DEC */
          CoordPair radecpos = poswcs->wcs()->
                          physicalCoords(sourcex, sourcey);
          ra  = radecpos.x();
          dec = radecpos.y();

	  errstr 
	    << "PositionUtils: From POS coordinates "
	    << " x=" << sourcex
	    << " y=" << sourcey
	    << " got "
	    << " ra= " << ra
	    << " dec= " << dec
	    << Sas::message(Msg::App, Msg::Noisy);
	}

     if ( (sctype == "EQPOS") || (sctype == "POS") )
     {
      /*** Now convert ra/dec to TELCOORDs */
      /* Code based on tool ccftools:srcinfov by U Lammers */

      // AttitudeMatrix A;
      SpacecraftAttitude SCA;

      if (poswcs->useODF()) 
      {
         oalServer.dataMode(Xmm::toDataMode(datamode));

        /* Get spacecraft attitude matrix from OAL */
         SCA = oalServer.attitudeLocator()->
                attitude(STime(dateobs)-Xmm::missionReferenceTime);
         // A = scAttitude.toAttitudeMatrix();
     
      }
      else 
      {
         //CoordPair radecpos = poswcs->wcs()->physicalRefCoords();
         SCA = SpacecraftAttitude( poswcs->refra() * DEG_TO_RAD,
                          poswcs->refdec() * DEG_TO_RAD,
                          poswcs->pa() * DEG_TO_RAD );

         // A = SpacecraftAttitude( radecpos.x() * DEG_TO_RAD, 
           //               radecpos.y() * DEG_TO_RAD,
            //              poswcs->pa() * DEG_TO_RAD ).toAttitudeMatrix();
      }

      errstr
                << "For datamode " << datamode
                << " and dateobs " << dateobs
                << " got an attitude matrix of \n" << SCA
                << Sas::message(Msg::App, Msg::Noisy);
     
      /* Get unit cartesian vector for source pointing in equatorial space */

      EquatorialDirection eqdirn( RightAscension(ra*DEG_TO_RAD),
                                     Declination(dec*DEG_TO_RAD) );
      CartesianVector srcVec = eqdirn.toUnitCartesian();

      errstr
                << "ra " << ra
                << " dec " << dec
                << " ra_rad " << ra*DEG_TO_RAD
                << " dec_rad " << dec*DEG_TO_RAD
                << " eqdirn " << eqdirn
                << " srcVec \n" << srcVec << "\n"
                << Sas::message(Msg::App, Msg::Noisy);

      //CartesianVector v0 = A * srcVec;
       
      /* Compute boresight matrix */
      Boresight * bs = 0; bs = calServer.getAtom(bs);

      //AttitudeMatrix BS = bs->matrix(0.);
      //AttitudeMatrix BSA = BS * A;
      //CartesianVector v = BSA * srcVec;
       
      bs->withSacCoordOrientation(SCA);
      XmmTelescopePointList pl = bs->projectOntoTelCoord(srcVec);
      theta=pl.theta()[0]; phi = pl.phi()[0];

      /* Get theta and phi */
      //theta = acos(v.x())*RAD_TO_DEG*3600.;
      //phi = fmod(3./2.*M_PI + atan2(v.z(), v.y()), M_TWO_PI);

     } 
     else        // TELcoords
     {
      theta = sourcex;
      phi = sourcey;
     }

    /* Finally convert theta/phi to DET coords */
     errstr 
	<< "PositionUtils: theta (arcsecs) =" << theta
	<< " phi (rad) = " << phi
	<< Sas::message(Msg::App, Msg::Noisy);


      GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);
      
      CalReal64Vector thetav(&theta, 1);
      CalReal64Vector phiv(&phi, 1);
  
      XmmTelescopePointList tpl = ct->telCoordList(thetav, phiv);

      XmmCartesianPointList cccpl = tpl.toCamCoord2();
//      XmmCartesianPointList cc1pl = cccpl.toCamCoord1();

      /* Get DET coordinate representation of source position */
      CoordPair sourceDet = CamCoord2Todetxy005(
                                  CamCoords(cccpl.x()[0], cccpl.y()[0], 1));
    
      //double sourcedetx = (cc1pl.x())[0] /_detToCamcoord2; // mm -> deg (DETX/Y)
      //double sourcedety = (cc1pl.y())[0] /_detToCamcoord2; // mm -> deg

      double sourcedetx = sourceDet.x();
      double sourcedety = sourceDet.y();

      errstr 
	<< "PositionUtils: detx= " << sourcedetx
	<< " dety= " << sourcedety
	<< Sas::message(Msg::App, Msg::Noisy);

  

      ALTRETURNV (SourcePosition( sourcedetx,
				  sourcedety,
				  theta,
				  phi ));

    }
  else if (sctype == "DET")
    {
      /* i/p coordinates are already in DET coords*/

      errstr 
	<< "[sourcepos] using DET coords [detx="
	<< sourcex
	<< ",dety="
	<< sourcey
	<<"]"
	<< Sas::message(Msg::App, Msg::Verbose);
      
      /* Get TELCOORDs representation of source position */
      CamCoords camSource = detxy005ToCamCoord2(sourcex, sourcey);
      TelCoords telSource = camCoord2ToTelCoord(camSource);

      errstr 
	<< "Got theta = " << telSource.theta()
	<< " and phi = " << telSource.phi()
	<< Sas::message(Msg::App, Msg::Noisy);

      ALTRETURNV( SourcePosition( sourcex, 
			     sourcey,
			     telSource.theta(), telSource.phi() ));
    }
  else if (sctype == "RAW")
    {
      // convert i/p coordinates from RAW to TelCoord
      /* Create PixCoord object from RAW position */
      PixCoords pixCoord1(UNT(sourcex), UNT(sourcey));
      CamCoords camSource = PixCoordTocamCoord2(pixCoord1, ccdnr);
      TelCoords telSource = camCoord2ToTelCoord(camSource);
      CoordPair detSource = CamCoord2Todetxy005(camSource);

      errstr
        << "Got theta = " << telSource.theta()
        << " and phi = " << telSource.phi()
        << Sas::message(Msg::App, Msg::Noisy);

      ALTRETURNV( SourcePosition( detSource.x(), detSource.y(),
                             telSource.theta(), telSource.phi() ));

      
    }

  errstr << "The input coordinate system [" 
	 << sourceCoordType
	 << "] is not valid. This should not have happened - please raise an SPR."
	 << Sas::error("sourceCoordTypeInvalid");
      
  RETURN SourcePosition (0,0,0,0);
  
}

ostream & operator<< (ostream & os, const PositionUtils & pu) {
	os << "_detToCamcoord2: " << pu._detToCamcoord2 << endl;
	os << "_cenDetX: " << pu._cenDetX << " _cenDetY: " << pu._cenDetY << endl;
	os << "X(" << pu._ccd_xmin << ", " << pu._ccd_ymax << ")\t";
	os <<"Y(" << pu._ccd_ymin << ", " << pu._ccd_ymax << ")" << endl;
	os << "_cent_ccd: " << pu._cent_ccd << " _nccds: " << pu._nccds << endl;

	return os;
}


