// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//      PsfRegion.h:       --- region for PSF class
//
//      DESCRIPTION:
//              An instance of class <PsfRegion> represents a geometrical area in
//              two dimensions that covers a specified PSF; it consists of one 
//              of the following elemental shapes:
//                      o circle
//                      o ellipse
//                      o box
//                      o rotbox
//
//	STRUCTURAL IDEA:
//		<PsfRegion> heritates from <GeoRegion> and <FBox>. Itself is region 
//		formed by the intersection of a list of constiuent <Component> which
//		itself is the union of a list of a list of constituent shapes. (see 
//		Region.h)


#ifndef PSFREGION_H
#define PSFREGION_H

#include <string>

#include "Region.h"	// Region.h has the defination of parent class <GeoRegion>

class PsfRegion : public GeoRegion {
public:
	// default constructor - create empty region
	PsfRegion() : GeoRegion(), _CoverBox(0) {}
	~PsfRegion() { if (_CoverBox) delete _CoverBox; }
	
	// construct region from string expression, i.e. (DETX,DETY) IN circle(132.5,-232.5,1843.58)
	PsfRegion(const string & expr) : GeoRegion() { setRegion(expr); }

	// utility functions:
	void setRegion(const string & expr) throw (InvalidRegion);	// set region with a string expression
	void setExpression(const string & expr = "");	// set region expression
	void setCoordType(const string & type); // set coordinate type
        bool setBox();          // set <FBox> properties according to the <GeoRegion> properties

	bool isEmpty() const { return (empty() || at(0).empty()); }
	const string & getExpression() const { return _Expression; }	// return the region expression
	const string & getCoordType() const { return _CoordType; } // return the coordinate type, i.e RAW, DET or POS
	ShapeType getRegionType() const { return _RegionType; } // return the shape of the region
	const FBox * getCoverBox() const { return _CoverBox; } // return the FBox object

private:
	void ellipseTrans(float rot, float major, float minor, float & Xmax, float & YMax); // given rotation angle and radius, return the max (x,y) after rotation
	void boxTrans(float rot, float x, float y, float & Xmax, float & Ymax);// given original (x,y) and width & height, return the max (x,y) after rotation

protected:
	FBox *	_CoverBox;
	string	_Expression;
	ShapeType _RegionType;
	string	_CoordType;
};

std::ostream & operator<< (std::ostream & out, const PsfRegion & reg);
#endif
