//
// Psfgen.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to write a set of PSFs into a FITS image
// 
//#include <ctime>
#include <unistd.h> // For getpid

#include "Tracer.h"
#include "errstr.h"
#include "param.h"

#include "CalServer.h"
#include "Dal.h"
#include "DssBuilder.h"
#include "DatafileAttributes.h"

#include "SscWcsConvertor.h"
#include "Psfgen.h"

/* Method: Psfgen() - Constructor */
Psfgen::Psfgen()
{
  PROC;

  /* Initialise the input image attributes */
    _atts=0;

  /* Get the parameters from this command line */
    string imagefile = stringParameter("image");
    string psffile = stringParameter("output");
    string instr = stringParameter("instrument");
    string slevel = uppercase(stringParameter("level"));
    AccuracyLevel level = getAccuracyLevel(slevel);

  // Get energies and weights
    unsigned nenergy = parameterCount("energy");
    vector<float> energy;
    for (unsigned i=0; i<nenergy; i++)
        energy.push_back(realParameter("energy", i));

  // Get weights normalised to 1
    vector<float> weights = getWeights(nenergy);

  /* Read the PSF position parameters */
    string coordtype = uppercase(stringParameter("coordtype"));
    double x = realParameter("x");
    double y = realParameter("y");

  /* Set the output image size */
    int xsize = intParameter("xsize");
    int ysize = intParameter("ysize");
    
    // Set the CCD if RAW coordinates have been selected
    if (coordtype == "RAW" && ! booleanParameter("withccd")) 
        errstr << "The CCD number (ccdnr) must be set if "
               << "position is given in RAW coordinates"
    	       << Sas::error("noCCDNumber");

  /* If an image has been given extract information from this */
    if (booleanParameter("withimage")) {

       /* Set the ssclib wcs and attitude structure */
        SscWcsConvertor::setWcs(imagefile);
        SscWcsConvertor::setWcsRef(imagefile);
        SscWcsConvertor::setAtt(imagefile);

       /* Get keywords from the image */
        _atts = new DatafileAttributes(imagefile);

       /* If instrument hasn't been set on the command line use value from image */
        if (!booleanParameter("withinstrument"))
           { instr = _atts->instrument(); }

    } else {

       /* If coordinates are POS or EQPOS we need an image to convert
          to TEL coords */
        if (coordtype == "POS" or coordtype == "EQPOS" or
            coordtype == "IMPIX")
           errstr << "An image must be supplied if PSF centre is given "
                  << "in X/Y, RA,DEC or Image Pixel coordinates"
                  << Sas::error("noImageSupplied");
    }

  /* Set the CAL state instrument and PSF mode */
    calServer.setInstrument(Xmm::toInstrument(instr));
    calServer.state()->accuracyLevel()->set(level);

    calServer.state()->ccd()->set(intParameter("ccdnr"));

  /* Find the PSF centre position in TEL coords (theta,phi) and set celestial position */
    CoordPair tel = findTelCoords(coordtype, x, y);

    errstr<< "Making PSF file "<<psffile<<" at theta="<<tel.x()
          <<" arcsec, phi="<<tel.y()*180.0/M_PI<<" degrees "
          <<"for instrument "<<instr
          <<Sas::message(Msg::App, Msg::Verbose);

  /* Set the rotation angle from parameter or from image PA angle */
    double rotate = 0.0;
    if (realParameter("rotate") > 0) {
       rotate = realParameter("rotate");
    } else if (booleanParameter("withimage")) {
       //rotate = 360.0 - _atts->papnt();
       rotate = _atts->papnt();
    }

    rotate = rotate * M_PI / 180.0; // Needed in radians

  /* Get the PSF from the CAL */
    PsfDataServer *pds = 0; pds = calServer.getAtom(pds);

  /* CAlculate window size in mm */
    double halfWindowXmm=0.0, halfWindowYmm=0.0;
    if (level == LOW || level == MEDIUM || level == HIGH) {
       halfWindowXmm = (xsize + 1) / 50.0; // 1.1" binned psf image
       halfWindowYmm = (ysize + 1) / 50.0;
    } else { // EXTENDED and ELLBETA PSFs have 1.0" bin images
       halfWindowXmm = xsize / 55.0;
       halfWindowYmm = ysize / 55.0;
    }

    // Loop over each energy
    vector<Image<float> > psfmap;
    for (unsigned i = 0; i<energy.size(); i++) {

       // Get the PSF for the selected position and energy
       Psf &p = pds->psf(energy[i], tel.x(), tel.y());
       Psf * psf = &p;

       // Rotate as required
       psf->setPosAngle(rotate);

       psfmap.push_back(psf->image(SpatialWindow<double>(halfWindowXmm,
                               halfWindowYmm)));
    }

  /* Write PSF(s) to the output file */
    writePsf(psfmap, weights, energy, psffile, xsize, ysize);

ENDPROC
}

/* Destructor */
Psfgen::~Psfgen()
{
  PROC;

    delete _atts; _atts=0;

  ENDPROC;
};

/* Method: findTelCoords - calculates the tel coords from any input */
CoordPair Psfgen::findTelCoords(string &coordtype, double x, double y)
{
    CoordPair det(x,y), raw(x,y), tel(x,y);
    CoordPair cel(x * DEG_TO_RAD,y * DEG_TO_RAD); // RA/DEC in radians

    if (coordtype == "TEL") {
       // Set above
    } else if (coordtype == "DET") {
       tel = SscWcsConvertor::detToTel(x,y);
    } else if (coordtype == "RAW") { // ccd already set in cal state
       det = SscWcsConvertor::rawToDet(x,y);      
       tel = SscWcsConvertor::detToTel(det.x(),det.y());
    } else if (coordtype == "POS") {
       cel = SscWcsConvertor::xyToRaDec(x,y);
       tel = SscWcsConvertor::raDecToTel(cel.x(), cel.y());
    } else if (coordtype == "EQPOS") {
       tel = SscWcsConvertor::raDecToTel(cel.x(), cel.y());
    } else if (coordtype == "IMPIX") {
       cel = SscWcsConvertor::imagePixelsToRaDec(x,y);
       tel = SscWcsConvertor::raDecToTel(cel.x(), cel.y());
    } else {
            errstr << "Coordtype unknown: "<<coordtype
             << Sas::error("invalidCoordType");
    }

    // Here we set the global variable for the celestial position if known (not nice).
    if ((coordtype == "DET")||(coordtype == "RAW")||(coordtype == "TEL")) 
       _celpos=CoordPair(0.0,0.0);
    else
       _celpos=CoordPair( cel.x()/DEG_TO_RAD, cel.y()/DEG_TO_RAD );
    
    return tel;
}

/* Method: writePsf - writes a PSF extension block */
void Psfgen::writePsf(vector<Image<float> > psfmap,
                      vector<float> weight, 
                      vector<float> energy, string &psffile,
                      int xsize, int ysize)
{

  /* Open the output file */
    DataSet *dataSet = dataSetServer->open(psffile, Dal::Create);

  // Write the data headers
    if (psfmap.size()>1) 
    {
       // When multiple energies have been entered the primary array 
       // is the weighted combination of each PSF
       Image<float> combimage(psfmap[0].xExtent(),psfmap[0].yExtent());

       for (unsigned ip=0; ip<psfmap.size(); ++ip) {
          for (unsigned iy=0; iy<psfmap[ip].yExtent(); ++iy) {
            for (unsigned ix=0; ix<psfmap[ip].xExtent(); ++ix) {
              combimage[ix][iy] += psfmap[ip][ix][iy]*weight[ip];
              //primearr[i++] += psfmap[ip][ix][iy]*weight[ip]
            }
          }
       }

       // Write the primary
       string extname = "PRIMARY", senergy = "Summation";
       makePsfExtension(dataSet, extname, combimage, senergy,xsize,ysize);

       // Add the individual PSFs as extensions
       for (unsigned ip=0; ip<psfmap.size(); ++ip) {
         
           senergy = toString(energy[ip]);
           extname = "PSF_" + senergy;

           makePsfExtension(dataSet, extname,psfmap[ip],senergy,xsize,ysize);
       }

    } else { // Only one PSF

       string senergy = toString(energy[0]);
       string extname = "PRIMARY";

       makePsfExtension(dataSet, extname, psfmap[0], senergy,xsize,ysize);
    }

  // Close down
  dataSetServer->close(dataSet);

}
// End of writePsf 

// Method to write a block containing a PSF at a given energy
void Psfgen::makePsfExtension(DataSet *dataSet, string &extname,
                              Image<float> psfimage, string &senergy,
                              int xsize, int ysize)
{

  /* Create the array data structure */
  std::vector<unsigned long> extents(2), realext(2);
  extents[0] = ysize; // Required size
  extents[1] = xsize; // Or should Y and X be swapped ?
  realext[0] = psfimage.yExtent(); // Size returned by CAL
  realext[1] = psfimage.xExtent(); // Or should Y and X be swapped ?

  Array *a = dataSet->addArray(extname, Dal::Real32, extents);
  ArrayData *ad = a->data();
  real32 *im = ad->real32Data();

  // Handle the fact that CAL can return larger image
  int startrow = (realext[0] - extents[0]) / 2;
  int startcol = (realext[1] - extents[1]) / 2;

  /* Write image to array */
  /* Original code writes X elements backwards for non XY output */
  int i=0;
  for (int iy=0; iy<ysize; ++iy) 
      for (int ix=0; ix<xsize; ++ix) 
          im[i++] = psfimage[startcol+ix][startrow+iy];

  // Add standard keywords to Primary header
  if (extname.find("PRIMARY") != string::npos)
                        addPrimaryKeys(a); 

  /* Write energy string */
  *(a->addAttribute("ENERGY", "energy value of the PSF")) = senergy;

  /* Write WCS keywords */
  addWcs(a, xsize, ysize);

  // Close down arrays
  delete ad; ad=0;

  ENDPROC;
}

/* Add WCS keywords */
void Psfgen::addWcs(Array *a, int xsize, int ysize)
{
  float pix_xcen = xsize / 2.0;
  float pix_ycen = ysize / 2.0;

  // Set as RA, DEC centred on 0,0 or real requested position if possible
  double ra_cen = _celpos.x();  
  double dec_cen= _celpos.y();

  // Set the pixel size
  double pix_size=0.0;
  switch (calServer.state()->accuracyLevel()->get())
  {
   case LOW:
   case MEDIUM: 
   case HIGH:
     pix_size=1.1 / 3600.0;
     break;
   case ELLBETA:
   case EXTENDED:
   case SLEW:
     pix_size=1.0 / 3600.0;
     break;
   //default:
     // Shouldn't happen
  }

  * a->addAttribute("CTYPE1", "WCS coordinate name") = "RA---TAN";
  * a->addAttribute("CRPIX1", "WCS reference pixel") = pix_xcen;
  * a->addAttribute("CRVAL1", "[deg] coord. at X axis ref. pixel") = ra_cen;
  * a->addAttribute("CUNIT1", "Physical units of X axis") = "deg";
  * a->addAttribute("CDELT1", "WCS pixel size") = -1.0 * pix_size;
  * a->addAttribute("CTYPE2", "WCS coordinate name") = "DEC--TAN";
  * a->addAttribute("CRPIX2", "WCS reference pixel") = pix_ycen;
  * a->addAttribute("CRVAL2", "[deg] coord. at X axis ref. pixel") = dec_cen;
  * a->addAttribute("CUNIT2", "Physical units of X axis") = "deg";
  * a->addAttribute("CDELT2", "WCS pixel size") = pix_size;

} // End of add header keywords
  
/* Add primary header specific keywords */
void Psfgen::addPrimaryKeys(Array *a)
{

   * a->addAttribute( "HDUCLASS",
                      "Format conforms to OGIP/GSFC conventions")
                      = "OGIP";
   * a->addAttribute( "HDUCLAS1",
                                "File contains an Image")
                                      = "IMAGE";
   * a->addAttribute( "HDUCLAS2",
                               "Gross Image")
                               = "TOTAL";
   * a->addAttribute( "HDUVERS1",
                                "Version of format")
                                      = "1.1.0";
}
   
/* Convert the accuracy level from string */
AccuracyLevel Psfgen::getAccuracyLevel(string &slevel)
{
  AccuracyLevel accuracy = ELLBETA;

  if (slevel == "HIGH")
            accuracy = HIGH;
  else if ( slevel == "MEDIUM")
            accuracy = MEDIUM;
  else if ( slevel == "EXTENDED")
            accuracy = EXTENDED;
  else if ( slevel == "ELLBETA")
            accuracy = ELLBETA;
  else if ( slevel == "LOW")
            accuracy = LOW;
  else if ( slevel == "SLEW")
            accuracy = SLEW;
  // Bad input handled by configuration file

  return accuracy;
}

/* Get the weights from the command line */
vector<float> Psfgen::getWeights(unsigned nenergy)
{
  unsigned nweights = parameterCount("weight");

  vector<float> weights;

  // If the number of energies and weights is consistent read from parameter
  if (nenergy == nweights)
    {
       float sum=0.0;
       for (int i=0; i<parameterCount("weight"); i++){
           weights.push_back(realParameter("weight",i));
           sum += realParameter("weight",i);
       }
       for (unsigned i=0; i<weights.size(); i++)
           weights[i]=weights[i]/sum;
      
     } else { // Inconsistent number of weights so just set all equal
       for (int i=0; i<parameterCount("energy"); i++)
           weights.push_back(1.0/nenergy);
     }

  return weights;
}

